from __future__ import annotations

import asyncio
import logging
import time
from logging import Handler, LogRecord
from sys import _getframe
from typing import TYPE_CHECKING

from loguru import logger
from typing_extensions import override

from utilities.datetime import duration_to_timedelta
from utilities.logging import LogLevel

if TYPE_CHECKING:
    from utilities.types import Duration


class InterceptHandler(Handler):
    """Handler for intercepting standard logging messages.

    https://github.com/Delgan/loguru#entirely-compatible-with-standard-logging
    """

    @override
    def emit(self, record: LogRecord) -> None:
        # Get corresponding Loguru level if it exists.
        try:  # pragma: no cover
            level = logger.level(record.levelname).name
        except ValueError:  # pragma: no cover
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = _getframe(6), 6  # pragma: no cover
        while (  # pragma: no cover
            frame and frame.f_code.co_filename == logging.__file__
        ):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(  # pragma: no cover
            level, record.getMessage()
        )


def logged_sleep_sync(
    duration: Duration, /, *, level: LogLevel = LogLevel.INFO, depth: int = 1
) -> None:
    """Log a sleep operation, synchronously."""
    timedelta = duration_to_timedelta(duration)
    logger.opt(depth=depth).log(
        level, "Sleeping for {timedelta}...", timedelta=timedelta
    )
    time.sleep(timedelta.total_seconds())


async def logged_sleep_async(
    duration: Duration, /, *, level: LogLevel = LogLevel.INFO, depth: int = 1
) -> None:
    """Log a sleep operation, asynchronously."""
    timedelta = duration_to_timedelta(duration)
    logger.opt(depth=depth).log(
        level, "Sleeping for {timedelta}...", timedelta=timedelta
    )
    await asyncio.sleep(timedelta.total_seconds())


__all__ = ["InterceptHandler", "logged_sleep_async", "logged_sleep_sync"]



from __future__ import annotations

from contextlib import contextmanager
from http.client import HTTPSConnection
from ipaddress import IPv4Address
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator


def get_public_ip(*, timeout: float | None = None) -> IPv4Address:
    """Get your public IP address."""
    with yield_connection("api.ipify.org", timeout=timeout) as conn:
        conn.request("GET", "/?format=text")
        response = conn.getresponse()
        address = response.read().decode("utf8")
    return IPv4Address(address)


@contextmanager
def yield_connection(
    host: str, /, *, timeout: float | None = None
) -> Iterator[HTTPSConnection]:
    """Yield an HTTP connection."""
    conn = HTTPSConnection(host, timeout=timeout)
    try:
        yield conn
    finally:
        conn.close()


__all__ = ["get_public_ip", "yield_connection"]



from __future__ import annotations

from slack_sdk.webhook import WebhookClient
from slack_sdk.webhook.async_client import AsyncWebhookClient

from utilities.slack_sdk import _get_client_async, _get_client_sync


class TestGetClient:
    def test_sync(self) -> None:
        client = _get_client_sync("url")
        assert isinstance(client, WebhookClient)

    def test_async(self) -> None:
        client = _get_client_async("url")
        assert isinstance(client, AsyncWebhookClient)



from __future__ import annotations

from pathlib import Path
from re import search

from pytest import raises

from utilities.pytest import skipif_windows
from utilities.subprocess import (
    GetShellOutputError,
    _address_already_in_use_pattern,
    get_shell_output,
)


class TestAddressAlreadyInUsePattern:
    def test_pattern(self) -> None:
        pattern = _address_already_in_use_pattern()
        text = "OSError: [Errno 98] Address already in use"
        assert search(pattern, text) is not None


class TestGetShellOutput:
    @skipif_windows
    def test_main(self) -> None:
        output = get_shell_output("ls")
        assert any(line == "pyproject.toml" for line in output.splitlines())

    @skipif_windows
    def test_activate(self, *, tmp_path: Path) -> None:
        venv = Path(tmp_path, ".venv")
        activate = Path(venv, "activate")
        activate.parent.mkdir(parents=True)
        activate.touch()
        _ = get_shell_output("ls", cwd=venv, activate=venv)

    def test_no_activate(self, *, tmp_path: Path) -> None:
        venv = Path(tmp_path, ".venv")
        with raises(GetShellOutputError):
            _ = get_shell_output("ls", cwd=venv, activate=venv)



from __future__ import annotations

from dataclasses import dataclass
from http import HTTPStatus

from slack_sdk.webhook import WebhookClient, WebhookResponse
from slack_sdk.webhook.async_client import AsyncWebhookClient
from typing_extensions import override

from utilities.functools import cache

_TIMEOUT = 30


def send_slack_sync(text: str, /, *, url: str, timeout: int = _TIMEOUT) -> None:
    """Send a message to Slack, synchronously."""
    client = _get_client_sync(url, timeout=timeout)  # pragma: no cover
    response = client.send(text=text)  # pragma: no cover
    _check_status_code(response)  # pragma: no cover


async def send_slack_async(
    text: str,
    /,
    *,
    url: str,
    timeout: int = _TIMEOUT,  # noqa: ASYNC109
) -> None:
    """Send a message via Slack."""
    client = _get_client_async(url, timeout=timeout)  # pragma: no cover
    response = await client.send(text=text)  # pragma: no cover
    _check_status_code(response)  # pragma: no cover


def _check_status_code(response: WebhookResponse, /) -> None:
    """Check that a chunk was successfully sent."""
    if response.status_code != HTTPStatus.OK:  # pragma: no cover
        raise SendSlackError(response=response)


@dataclass(kw_only=True)
class SendSlackError(Exception):
    response: WebhookResponse

    @override
    def __str__(self) -> str:
        return f"Webhook response was not OK; got {self.response.status_code}"  # pragma: no cover


@cache
def _get_client_sync(url: str, /, *, timeout: int = _TIMEOUT) -> WebhookClient:
    """Get the webhook client."""
    return WebhookClient(url, timeout=timeout)


@cache
def _get_client_async(url: str, /, *, timeout: int = _TIMEOUT) -> AsyncWebhookClient:
    """Get the engine/sessionmaker for the required database."""
    return AsyncWebhookClient(url, timeout=timeout)


__all__ = ["send_slack_async", "send_slack_sync"]
